import typing as t
from discord.ext import commands

class Router:
    """ Create a router, connected to a bot instance to allow route-based command registry.

        :Example Usage:
        .. code-block:: python

            bot = commands.Bot()
            router = Router(bot)

            @router.route(["cmd", "subcmd"])
            async def subcmd(ctx):
                pass
    """

    def __init__(self, bot: commands.Bot):
        self.bot = bot

    def route(self, command: list, **kwargs):
        """ A shortcut decorator that registers a command route.

            :param list[str] command: A list of strings defining the route to the command.
            :param **kwargs: Any command attributes to pass on, such as aliases.
        """
        def deco(func):
            return self.get_command_group(command, func, **kwargs)
        return deco

    def get_command_group(self, path: list, func: t.Optional[t.Callable] = None, **kwargs) -> commands.Group:
        # Return self.bot if path is the root
        if len(path) == 0:
            return self.bot
        # Try and find group
        group: commands.Group = self.bot.get_command(" ".join(path))
        # If it doesn't exist, create it group
        if group is None or func:
            # Get the parent group
            parent = self.get_command_group(path[:-1])
            existing_subcommands = []
            # Create the func if it doesn't exist
            if func is None:
                async def func(ctx):
                    raise commands.CommandNotFound(f'Command "{path[-1]}" is not found')
            # Remove the current group it already exists
            elif group is not None:
                existing_subcommands = group.commands
                parent.remove_command(group.name)

            # Create the group
            group = commands.Group(func, name=path[-1], invoke_without_command=True, **kwargs)
            # Add existing subcommands
            for cmd in existing_subcommands:
                group.add_command(cmd)
            # Register command
            parent.add_command(group)

        return group



"""Contains tools to help with managing times."""

import re
from datetime import datetime

from discord.ext import commands
from dateutil.relativedelta import relativedelta


class DurationConverter(commands.Converter):
    """Converts to a duration of time."""

    symbols = """
    - years: `y/Y`, `yr(s)`, `year(s)`
    - months: `m`, `mon(s)`, `month(s)`
    - weeks: `w/W`, `week(s)`
    - days: `d/D`, `day(s)`
    - hours: `h/H`, `hr(s)`, `hour(s)`
    - minutes: `M`, `min(s)`, `minute(s)`
    - seconds: `s/S`, `sec(s)`, `second(s)`

    Units must be provided in descending order of magnitude.
    """

    complied = re.compile(
        r"((?P<years>\d+?) ?(years|year|yrs|yr|Y|y) ?)?"
        r"((?P<months>\d+?) ?(months|month|mons|mon|m) ?)?"
        r"((?P<weeks>\d+?) ?(weeks|week|W|w) ?)?"
        r"((?P<days>\d+?) ?(days|day|D|d) ?)?"
        r"((?P<hours>\d+?) ?(hours|hour|hrs|hr|H|h) ?)?"
        r"((?P<minutes>\d+?) ?(minutes|minute|mins|min|M) ?)?"
        r"((?P<seconds>\d+?) ?(seconds|second|secs|sec|S|s))?"
    )

    async def convert(self, ctx, argument: str):
        match = self.complied.fullmatch(argument)
        if not match:
            return False

        delta = relativedelta(**{unit: int(amount) for unit, amount in match.groupdict(default=0).items()})
        now = datetime.utcnow()
        return now + delta



import asyncio
import discord

import blurple.ui as ui
import blurple.io as io


class ToastInteraction(io.ReactionAddBasic):
    """A custom toast interaction reply subclass."""

    async def on_reply_init(self, message: discord.Message):
        await super().on_reply_init(message)
        self.dismiss = "\u2716\ufe0f"
        self.validate = [self.dismiss]
        await self.message.add_reaction(self.dismiss)

    def reply_check(self, payload: discord.RawReactionActionEvent):
        return super().reply_check(payload) and payload.user_id == self.ctx.author.id

    async def on_reply_complete(self):
        await self.message.delete()


class Toast(ui.Base):
    """ A subclass of :class:`discord.Embed` for stylish toasts.

        The difference between this class and :class:`Alert` is these are intended to hold more deemphasized information.
        These are used with :class:`blurple.io.Toast`, which supports a "dismiss" button to dismiss the toast, and automatic deletion with specified duration.

        :param Style style: The style of the toasts.
        :param str text: The text of the toasts, will be wrapped in an emoji unless specified in options.
        :param **options: Toast options to customize it's look.

            :emoji: Defaults to :class:`True`. Can be set to false to remove the emoji from the toast.
    """

    def __init__(self, style: ui.Style, text: str, **options):
        super().__init__(
            color=style[0],
            description=self.process_text(style, text, **options),
        )

    async def send(self, client: discord.abc.Messageable, duration: int = 5):
        """ Send the toast out.

            :param client: The client used, usually a :class:`discord.abc.Messageable`. Must have implemented :func:`.send`
            :param int duration: The duration the toast lasts before disappearing in seconds
        """
        message = await client.send(embed=self)
        asyncio.create_task(ToastInteraction(client, message=message, timeout=duration).result())

    @classmethod
    def process_text(cls, style: ui.Style, text: str, **options):
        output: str = ''

        if options.get("emoji") is not False:
            output += style[1] + " "

        return output + text



from enum import Enum


class Style(Enum):
    """ A contextual class to style components.

    There are 9 `main styles <https://cdn.discordapp.com/attachments/598870131182927873/814573079371448330/unknown.png>`_.

    - ``PRIMARY``
    - ``SECONDARY``
    - ``SUCCESS``
    - ``DANGER``
    - ``WARNING``
    - ``INFO``
    - ``LIGHT``
    - ``DARK``
    - ``GHOST``

    You can also create a `custom style <https://cdn.discordapp.com/attachments/598870131182927873/811087938678685707/unknown.png>`_:

    :Example Usage:
    .. code-block:: python

        grape_style = (0x9266CC, "\\U0001f347", "Grape")

    ``PRIMARY`` and ``SECONDARY`` styles use custom emoji, so are unable to be used out of the box.
    To work around this, I've provided the source .svgs in the repository for the custom emojis used throughout the project. you can add these to a server that your bot is in, then create a custom style.

    Alternatively, if you want, you can support me on `ko-fi <https://ko-fi.com/s/7705c20532>`_, and I'll invite your bot to my server with the original custom emojis.

    """
    PRIMARY = (0x7289DA, "<:primary:808874731763007488>", "Primary")
    SECONDARY = (0x99AAB5, "<:secondary:808874731758813205>", "Secondary")
    SUCCESS = (0x77B255, "\u2705", "Success")
    DANGER = (0xDD2E44, "\U0001f6ab", "Danger")
    WARNING = (0xFFCC4D, "\u26a0\ufe0f", "Warning")
    INFO = (0x3B88C3, "\u2139\ufe0f", "Info")
    LIGHT = (0xE6E7E8, "\U0001f533", "Light")
    DARK = (0x31373D, "\U0001f532", "Dark")
    GHOST = (0x2f3136, "", "")

    def __getitem__(self, key):
        return self.value[key]



import discord

import blurple.io as io


class ReactionAddBasic(io.Reply):
    """An unopinionated, lower level class to wait for a user to add a reaction."""
    event = "raw_reaction_add"

    async def on_reply_init(self, message: discord.Message):
        """Sepcialized to pass message object."""
        self.message = message

    def reply_check(self, payload: discord.RawReactionActionEvent):
        """Specialized to check if the reaction and payload message is valid."""
        if payload.message_id == self.message.id and not payload.user_id == self.ctx.me.id:
            if self._iscontainer(self.validate):
                return str(payload.emoji) in self.validate
            return True


class ReactionRemoveBasic(ReactionAddBasic):
    """An unopinionated, lower level class to wait for a user to remove a reaction."""
    event = "raw_reaction_remove"


class ReactionAddReply(ReactionAddBasic):
    """ Ask for the user's reaction reply.

        :Example Usage:
        .. code-block:: python

            reply = await io.ReactionAddBasic(ctx, validate=["✅", "❎"]).result()
    """

    async def on_reply_init(self, message: discord.Message):
        """Specialized to add vaild reaction emojis to message, if validation is on."""
        await super().on_reply_init(message)
        if self._iscontainer(self.validate):
            for react in self.validate:
                await self.message.add_reaction(react)

    def reply_check(self, payload: discord.RawReactionActionEvent):
        """Specialized to check if payload user and message are valid."""
        return payload.user_id == self.ctx.author.id and \
               payload.message_id == self.message.id

    async def on_reply_attempt(self, payload: discord.RawReactionActionEvent):
        """Specialized to remove the user's reaction."""
        await self.message.remove_reaction(payload.emoji, self.ctx.bot.get_user(payload.user_id))
        return payload

    async def on_reply_complete(self):
        """Specialized to clear all reactions off the message."""
        await self.message.clear_reactions()



from abc import ABC
import discord


class Base(discord.Embed, ABC):

    async def send(self, client: discord.abc.Messageable):
        """ Send the component as a message in discord.

        :param client: The client used, usually a :class:`discord.abc.Messageable`. Must have implemented :func:`.send`
        :returns: :class:`discord.Message`
        """
        return await client.send(embed=self)
